@@ -1,4 +1,4 @@
-# Copyright 2016-2020, Optimizely
+# Copyright 2016-2021, Optimizely
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
 # You may obtain a copy of the License at
@@ -10,6 +10,7 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+
 from six import string_types
 
 from . import decision_service
@@ -20,31 +21,37 @@ from . import logger as _logging
 from .config_manager import AuthDatafilePollingConfigManager
 from .config_manager import PollingConfigManager
 from .config_manager import StaticConfigManager
+from .decision.optimizely_decide_option import OptimizelyDecideOption
+from .decision.optimizely_decision import OptimizelyDecision
+from .decision.optimizely_decision_message import OptimizelyDecisionMessage
 from .error_handler import NoOpErrorHandler as noop_error_handler
 from .event import event_factory, user_event_factory
 from .event.event_processor import ForwardingEventProcessor
 from .event_dispatcher import EventDispatcher as default_event_dispatcher
 from .helpers import enums, validator
+from .helpers.enums import DecisionSources
 from .notification_center import NotificationCenter
 from .optimizely_config import OptimizelyConfigService
+from .optimizely_user_context import OptimizelyUserContext
 
 
 class Optimizely(object):
     """ Class encapsulating all SDK functionality. """
 
     def __init__(
-        self,
-        datafile=None,
-        event_dispatcher=None,
-        logger=None,
-        error_handler=None,
-        skip_json_validation=False,
-        user_profile_service=None,
-        sdk_key=None,
-        config_manager=None,
-        notification_center=None,
-        event_processor=None,
-        datafile_access_token=None,
+            self,
+            datafile=None,
+            event_dispatcher=None,
+            logger=None,
+            error_handler=None,
+            skip_json_validation=False,
+            user_profile_service=None,
+            sdk_key=None,
+            config_manager=None,
+            notification_center=None,
+            event_processor=None,
+            datafile_access_token=None,
+            default_decide_options=None
     ):
         """ Optimizely init method for managing Custom projects.
 
@@ -68,6 +75,7 @@ class Optimizely(object):
                        which simply forwards events to the event dispatcher.
                        To enable event batching configure and use optimizely.event.event_processor.BatchEventProcessor.
       datafile_access_token: Optional string used to fetch authenticated datafile for a secure project environment.
+      default_decide_options: Optional list of decide options used with the decide APIs.
     """
         self.logger_name = '.'.join([__name__, self.__class__.__name__])
         self.is_valid = True
@@ -80,6 +88,17 @@ class Optimizely(object):
             self.event_dispatcher, logger=self.logger, notification_center=self.notification_center,
         )
 
+        if default_decide_options is None:
+            self.default_decide_options = []
+        else:
+            self.default_decide_options = default_decide_options
+
+        if isinstance(self.default_decide_options, list):
+            self.default_decide_options = self.default_decide_options[:]
+        else:
+            self.logger.debug('Provided default decide options is not a list.')
+            self.default_decide_options = []
+
         try:
             self._validate_instantiation_options()
         except exceptions.InvalidInputException as error:
@@ -192,7 +211,7 @@ class Optimizely(object):
             )
 
     def _get_feature_variable_for_type(
-        self, project_config, feature_key, variable_key, variable_type, user_id, attributes,
+            self, project_config, feature_key, variable_key, variable_type, user_id, attributes
     ):
         """ Helper method to determine value for a certain variable attached to a feature flag based on type of variable.
 
@@ -245,7 +264,7 @@ class Optimizely(object):
         feature_enabled = False
         source_info = {}
         variable_value = variable.defaultValue
-        decision = self.decision_service.get_variation_for_feature(project_config, feature_flag, user_id, attributes)
+        decision, _ = self.decision_service.get_variation_for_feature(project_config, feature_flag, user_id, attributes)
         if decision.variation:
 
             feature_enabled = decision.variation.featureEnabled
@@ -328,7 +347,8 @@ class Optimizely(object):
         feature_enabled = False
         source_info = {}
 
-        decision = self.decision_service.get_variation_for_feature(project_config, feature_flag, user_id, attributes)
+        decision, _ = self.decision_service.get_variation_for_feature(
+            project_config, feature_flag, user_id, attributes)
         if decision.variation:
 
             feature_enabled = decision.variation.featureEnabled
@@ -520,7 +540,7 @@ class Optimizely(object):
         if not self._validate_user_inputs(attributes):
             return None
 
-        variation = self.decision_service.get_variation(project_config, experiment, user_id, attributes)
+        variation, _ = self.decision_service.get_variation(project_config, experiment, user_id, attributes)
         if variation:
             variation_key = variation.key
 
@@ -577,7 +597,7 @@ class Optimizely(object):
 
         feature_enabled = False
         source_info = {}
-        decision = self.decision_service.get_variation_for_feature(project_config, feature, user_id, attributes)
+        decision, _ = self.decision_service.get_variation_for_feature(project_config, feature, user_id, attributes)
         is_source_experiment = decision.source == enums.DecisionSources.FEATURE_TEST
         is_source_rollout = decision.source == enums.DecisionSources.ROLLOUT
 
@@ -889,7 +909,7 @@ class Optimizely(object):
             self.logger.error(enums.Errors.INVALID_PROJECT_CONFIG.format('get_forced_variation'))
             return None
 
-        forced_variation = self.decision_service.get_forced_variation(project_config, experiment_key, user_id)
+        forced_variation, _ = self.decision_service.get_forced_variation(project_config, experiment_key, user_id)
         return forced_variation.key if forced_variation else None
 
     def get_optimizely_config(self):
@@ -913,3 +933,229 @@ class Optimizely(object):
             return self.config_manager.optimizely_config
 
         return OptimizelyConfigService(project_config).get_config()
+
+    def create_user_context(self, user_id, attributes=None):
+        """
+        We do not check for is_valid here as a user context can be created successfully
+        even when the SDK is not fully configured.
+
+        Args:
+            user_id: string to use as user id for user context
+            attributes: dictionary of attributes or None
+
+        Returns:
+            UserContext instance or None if the user id or attributes are invalid.
+        """
+        if not isinstance(user_id, string_types):
+            self.logger.error(enums.Errors.INVALID_INPUT.format('user_id'))
+            return None
+
+        if attributes is not None and type(attributes) is not dict:
+            self.logger.error(enums.Errors.INVALID_INPUT.format('attributes'))
+            return None
+
+        return OptimizelyUserContext(self, user_id, attributes)
+
+    def _decide(self, user_context, key, decide_options=None):
+        """
+        decide calls optimizely decide with feature key provided
+        Args:
+            user_context: UserContent with userid and attributes
+            key: feature key
+            decide_options: list of OptimizelyDecideOption
+
+        Returns:
+            Decision object
+        """
+
+        # raising on user context as it is internal and not provided directly by the user.
+        if not isinstance(user_context, OptimizelyUserContext):
+            raise exceptions.InvalidInputException(enums.Errors.INVALID_INPUT.format('user_context'))
+
+        reasons = []
+
+        # check if SDK is ready
+        if not self.is_valid:
+            self.logger.error(enums.Errors.INVALID_OPTIMIZELY.format('decide'))
+            reasons.append(OptimizelyDecisionMessage.SDK_NOT_READY)
+            return OptimizelyDecision(flag_key=key, user_context=user_context, reasons=reasons)
+
+        # validate that key is a string
+        if not isinstance(key, string_types):
+            self.logger.error('Key parameter is invalid')
+            reasons.append(OptimizelyDecisionMessage.FLAG_KEY_INVALID.format(key))
+            return OptimizelyDecision(flag_key=key, user_context=user_context, reasons=reasons)
+
+        # validate that key maps to a feature flag
+        config = self.config_manager.get_config()
+        if not config:
+            self.logger.error(enums.Errors.INVALID_PROJECT_CONFIG.format('decide'))
+            reasons.append(OptimizelyDecisionMessage.SDK_NOT_READY)
+            return OptimizelyDecision(flag_key=key, user_context=user_context, reasons=reasons)
+
+        feature_flag = config.get_feature_from_key(key)
+        if feature_flag is None:
+            self.logger.error("No feature flag was found for key '#{key}'.")
+            reasons.append(OptimizelyDecisionMessage.FLAG_KEY_INVALID.format(key))
+            return OptimizelyDecision(flag_key=key, user_context=user_context, reasons=reasons)
+
+        # merge decide_options and default_decide_options
+        if isinstance(decide_options, list):
+            decide_options += self.default_decide_options
+        else:
+            self.logger.debug('Provided decide options is not an array. Using default decide options.')
+            decide_options = self.default_decide_options
+
+        # Create Optimizely Decision Result.
+        user_id = user_context.user_id
+        attributes = user_context.get_user_attributes()
+        variation_key = None
+        variation = None
+        feature_enabled = False
+        rule_key = None
+        flag_key = key
+        all_variables = {}
+        experiment = None
+        decision_source = DecisionSources.ROLLOUT
+        source_info = {}
+        decision_event_dispatched = False
+        ignore_ups = OptimizelyDecideOption.IGNORE_USER_PROFILE_SERVICE in decide_options
+
+        decision, decision_reasons = self.decision_service.get_variation_for_feature(config, feature_flag, user_id,
+                                                                                     attributes, ignore_ups)
+
+        reasons += decision_reasons
+
+        # Fill in experiment and variation if returned (rollouts can have featureEnabled variables as well.)
+        if decision.experiment is not None:
+            experiment = decision.experiment
+            source_info["experiment"] = experiment
+            rule_key = experiment.key
+        if decision.variation is not None:
+            variation = decision.variation
+            variation_key = variation.key
+            feature_enabled = variation.featureEnabled
+            decision_source = decision.source
+            source_info["variation"] = variation
+
+        # Send impression event if Decision came from a feature
+        # test and decide options doesn't include disableDecisionEvent
+        if OptimizelyDecideOption.DISABLE_DECISION_EVENT not in decide_options:
+            if decision_source == DecisionSources.FEATURE_TEST or config.send_flag_decisions:
+                self._send_impression_event(config, experiment, variation, flag_key, rule_key or '',
+                                            decision_source, feature_enabled,
+                                            user_id, attributes)
+                decision_event_dispatched = True
+
+        # Generate all variables map if decide options doesn't include excludeVariables
+        if OptimizelyDecideOption.EXCLUDE_VARIABLES not in decide_options:
+            for variable_key in feature_flag.variables:
+                variable = config.get_variable_for_feature(flag_key, variable_key)
+                variable_value = variable.defaultValue
+                if feature_enabled:
+                    variable_value = config.get_variable_value_for_variation(variable, decision.variation)
+                    self.logger.debug(
+                        'Got variable value "%s" for variable "%s" of feature flag "%s".'
+                        % (variable_value, variable_key, flag_key)
+                    )
+
+                try:
+                    actual_value = config.get_typecast_value(variable_value, variable.type)
+                except:
+                    self.logger.error('Unable to cast value. Returning None.')
+                    actual_value = None
+
+                all_variables[variable_key] = actual_value
+
+        should_include_reasons = OptimizelyDecideOption.INCLUDE_REASONS in decide_options
+
+        # Send notification
+        self.notification_center.send_notifications(
+            enums.NotificationTypes.DECISION,
+            enums.DecisionNotificationTypes.FLAG,
+            user_id,
+            attributes or {},
+            {
+                'flag_key': flag_key,
+                'enabled': feature_enabled,
+                'variables': all_variables,
+                'variation_key': variation_key,
+                'rule_key': rule_key,
+                'reasons': reasons if should_include_reasons else [],
+                'decision_event_dispatched': decision_event_dispatched
+
+            },
+        )
+
+        return OptimizelyDecision(variation_key=variation_key, enabled=feature_enabled, variables=all_variables,
+                                  rule_key=rule_key, flag_key=flag_key,
+                                  user_context=user_context, reasons=reasons if should_include_reasons else []
+                                  )
+
+    def _decide_all(self, user_context, decide_options=None):
+        """
+        decide_all will return a decision for every feature key in the current config
+        Args:
+            user_context: UserContent object
+            decide_options: Array of DecisionOption
+
+        Returns:
+            A dictionary of feature key to Decision
+        """
+        # raising on user context as it is internal and not provided directly by the user.
+        if not isinstance(user_context, OptimizelyUserContext):
+            raise exceptions.InvalidInputException(enums.Errors.INVALID_INPUT.format('user_context'))
+
+        # check if SDK is ready
+        if not self.is_valid:
+            self.logger.error(enums.Errors.INVALID_OPTIMIZELY.format('decide_all'))
+            return {}
+
+        config = self.config_manager.get_config()
+        if not config:
+            self.logger.error(enums.Errors.INVALID_PROJECT_CONFIG.format('decide'))
+            return {}
+
+        keys = []
+        for f in config.feature_flags:
+            keys.append(f['key'])
+        return self._decide_for_keys(user_context, keys, decide_options)
+
+    def _decide_for_keys(self, user_context, keys, decide_options=None):
+        """
+
+        Args:
+            user_context: UserContent
+            keys: list of feature keys to run decide on.
+            decide_options: an array of DecisionOption objects
+
+        Returns:
+            An dictionary of feature key to Decision
+        """
+        # raising on user context as it is internal and not provided directly by the user.
+        if not isinstance(user_context, OptimizelyUserContext):
+            raise exceptions.InvalidInputException(enums.Errors.INVALID_INPUT.format('user_context'))
+
+        # check if SDK is ready
+        if not self.is_valid:
+            self.logger.error(enums.Errors.INVALID_OPTIMIZELY.format('decide_for_keys'))
+            return {}
+
+        # merge decide_options and default_decide_options
+        merged_decide_options = []
+        if isinstance(decide_options, list):
+            merged_decide_options = decide_options[:]
+            merged_decide_options += self.default_decide_options
+        else:
+            self.logger.debug('Provided decide options is not an array. Using default decide options.')
+            merged_decide_options = self.default_decide_options
+
+        enabled_flags_only = OptimizelyDecideOption.ENABLED_FLAGS_ONLY in merged_decide_options
+
+        decisions = {}
+        for key in keys:
+            decision = self._decide(user_context, key, decide_options)
+            if enabled_flags_only and not decision.enabled:
+                continue
+            decisions[key] = decision
+        return decisions
